from __future__ import annotations

import shutil
import subprocess
from pathlib import Path

VERIFIED_BADGES_REPO_URL = "https://github.com/BioDisCo/verified-badges.git"


def ensure_verified_badges_repo(cache_root: Path, provided_repo: Path | None = None) -> Path:
    if provided_repo is not None:
        return provided_repo
    repo_dir = cache_root / "verified-badges"
    if repo_dir.exists():
        return repo_dir
    repo_dir.parent.mkdir(parents=True, exist_ok=True)
    subprocess.run(
        ["git", "clone", VERIFIED_BADGES_REPO_URL, str(repo_dir)],
        check=True,
        capture_output=True,
        text=True,
    )
    return repo_dir


def copy_verified_badges_assets(repo_dir: Path, paper_dir: Path) -> None:
    shutil.copy2(repo_dir / "verified-badges.sty", paper_dir / "verified-badges.sty")
    src_badges = repo_dir / "badges"
    dst_badges = paper_dir / "badges"
    if dst_badges.exists():
        shutil.rmtree(dst_badges)
    shutil.copytree(src_badges, dst_badges)


def infer_repo_url_and_branch(start_dir: Path) -> tuple[str | None, str | None]:
    root = _git_stdout(["git", "rev-parse", "--show-toplevel"], start_dir)
    if root is None:
        return None, None
    root_path = Path(root)
    remote = _git_stdout(["git", "remote", "get-url", "origin"], root_path)
    branch = _git_stdout(["git", "rev-parse", "--abbrev-ref", "HEAD"], root_path)
    if remote is None:
        return None, branch
    return _normalize_github_remote(remote), branch


def build_blob_url(repo_url: str | None, branch: str | None, absolute_path: Path) -> str | None:
    if repo_url is None or branch is None:
        return None
    root = _git_stdout(["git", "rev-parse", "--show-toplevel"], absolute_path.parent)
    if root is None:
        return None
    relative = absolute_path.resolve().relative_to(Path(root).resolve()).as_posix()
    return f"{repo_url}/blob/{branch}/{relative}"


def build_local_file_target(paper_dir: Path, absolute_path: Path) -> str:  # noqa: ARG001
    return absolute_path.resolve().as_uri()


def _git_stdout(cmd: list[str], cwd: Path) -> str | None:
    result = subprocess.run(
        cmd,
        cwd=cwd,
        text=True,
        capture_output=True,
        check=False,
    )
    if result.returncode != 0:
        return None
    return result.stdout.strip() or None


def _normalize_github_remote(remote: str) -> str:
    trimmed = remote.strip()
    if trimmed.endswith(".git"):
        trimmed = trimmed[:-4]
    if trimmed.startswith("git@github.com:"):
        return "https://github.com/" + trimmed.split(":", 1)[1]
    return trimmed



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any


@dataclass
class StatementTask:
    task_id: str
    tex_path: str
    env_name: str
    label: str
    title: str | None
    sequence_index: int
    kind: str
    source_excerpt: str
    status: str = "pending"
    attempts: int = 0
    lean_path: str | None = None
    badge_kind: str | None = None
    summary: str | None = None
    weeks_scale_reason: str | None = None
    last_error: str | None = None
    attempt_history: list[str] = field(default_factory=lambda: list[str]())  # noqa: PLW0108
    transport_failures: int = 0

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> StatementTask:
        return cls(**data)


@dataclass
class RunState:
    arxiv_id: str
    run_dir: str
    paper_dir: str
    main_tex: str
    repo_url: str | None
    branch: str | None
    prover_backend: str
    verifier_backend: str
    badge_link_mode: str = "auto"
    explicit_repo_target: bool = False
    is_local_paper: bool = False
    root_module: str | None = None
    tasks: list[StatementTask] = field(default_factory=lambda: list[StatementTask]())  # noqa: PLW0108

    def to_dict(self) -> dict[str, Any]:
        data = asdict(self)
        data["tasks"] = [task.to_dict() for task in self.tasks]
        return data

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> RunState:
        data = dict(data)
        data.setdefault("root_module", None)
        data.setdefault("explicit_repo_target", False)
        data.setdefault("prover_backend", "codex")
        data.setdefault("verifier_backend", "claude")
        data["tasks"] = [StatementTask.from_dict(item) for item in data.get("tasks", [])]
        return cls(**data)

    @property
    def state_path(self) -> Path:
        return Path(self.run_dir) / "state.json"



import json
from pathlib import Path

from paper_badger.dashboard import load_state, render_dashboard
from paper_badger.models import RunState, StatementTask


def test_render_dashboard_includes_summary_and_current_task(tmp_path: Path) -> None:
    run_dir = tmp_path / "paper-id"
    run_dir.mkdir()
    state = RunState(
        arxiv_id="paper-id",
        run_dir=str(run_dir),
        paper_dir=str(run_dir / "paper"),
        main_tex=str(run_dir / "paper" / "main.tex"),
        repo_url=None,
        branch=None,
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="local",
        root_module="Paper",
        tasks=[
            StatementTask(
                task_id="main.tex::0::lem:main",
                tex_path="main.tex",
                env_name="lemma",
                label="lem:main",
                title=None,
                sequence_index=0,
                kind="lemma",
                source_excerpt="If x = y then y = x.",
                status="proving",
                attempts=2,
                attempt_history=["Attempt 2: retry requested."],
            ),
            StatementTask(
                task_id="main.tex::1::thm:done",
                tex_path="main.tex",
                env_name="theorem",
                label="thm:done",
                title=None,
                sequence_index=1,
                kind="theorem",
                source_excerpt="Done theorem.",
                status="verified",
            ),
        ],
    )

    output = render_dashboard(state)

    assert "Run: paper-id" in output
    assert "1/2 completed (1 verified, 0 formalized)" in output
    assert "- lemma lem:main [proving]" in output
    assert "- lem:main [proving] attempts=2" in output
    assert "- [!] `lemma` `lem:main`: in progress" in output
    assert "- [x] `theorem` `thm:done`: verified (verified)" in output


def test_load_state_reads_state_json(tmp_path: Path) -> None:
    run_dir = tmp_path / "paper-id"
    run_dir.mkdir()
    state = RunState(
        arxiv_id="paper-id",
        run_dir=str(run_dir),
        paper_dir=str(run_dir / "paper"),
        main_tex=str(run_dir / "paper" / "main.tex"),
        repo_url=None,
        branch=None,
        prover_backend="codex",
        verifier_backend="codex",
        tasks=[],
    )
    (run_dir / "state.json").write_text(json.dumps(state.to_dict()), encoding="utf-8")

    loaded = load_state(run_dir)

    assert loaded.arxiv_id == "paper-id"



from __future__ import annotations

import json
import shutil
import subprocess
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .latex import sanitize_label_stem, statement_context_summary

if TYPE_CHECKING:
    from .models import StatementTask

PROVER_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "status",
        "summary",
        "lean_path",
        "proof_complete",
        "weeks_scale",
        "weeks_scale_reason",
        "paper_issue",
    ],
    "properties": {
        "status": {
            "type": "string",
            "enum": [
                "verified_candidate",
                "formalized_candidate",
                "needs_retry",
                "weeks_scale",
            ],
        },
        "summary": {"type": "string"},
        "lean_path": {"type": ["string", "null"]},
        "proof_complete": {"type": "boolean"},
        "weeks_scale": {"type": "boolean"},
        "weeks_scale_reason": {"type": ["string", "null"]},
        "paper_issue": {"type": ["string", "null"]},
    },
}

VERIFIER_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": [
        "decision",
        "summary",
        "exact_match",
        "extra_hypotheses",
        "weeks_scale_reason",
    ],
    "properties": {
        "decision": {
            "type": "string",
            "enum": ["verified", "formalized", "retry", "weeks_scale"],
        },
        "summary": {"type": "string"},
        "exact_match": {"type": "boolean"},
        "extra_hypotheses": {"type": "boolean"},
        "weeks_scale_reason": {"type": ["string", "null"]},
    },
}


@dataclass
class ProverResult:
    status: str
    summary: str
    lean_path: str | None
    proof_complete: bool
    weeks_scale: bool
    weeks_scale_reason: str | None = None
    paper_issue: str | None = None


@dataclass
class VerifierResult:
    decision: str
    summary: str
    exact_match: bool
    extra_hypotheses: bool
    weeks_scale_reason: str | None = None


class AgentTimeoutError(RuntimeError):
    """Raised when a prover or verifier call exceeds the configured timeout."""


class AgentInvocationError(RuntimeError):
    """Invocation failed before returning valid structured output."""


class AgentTransportError(AgentInvocationError):
    """Backend connection failed before returning structured output."""


class AgentRunner:
    def __init__(
        self,
        run_dir: Path,
        paper_dir: Path,
        main_tex: Path,
        root_module: str | None,
        prover_backend: str,
        verifier_backend: str,
        agent_timeout_seconds: int | None = None,
        codex_model: str | None = None,
        claude_model: str | None = None,
    ) -> None:
        self.run_dir = run_dir
        self.paper_dir = paper_dir
        self.main_tex = main_tex
        self.root_module = root_module
        self.prover_backend = prover_backend
        self.verifier_backend = verifier_backend
        self.agent_timeout_seconds = agent_timeout_seconds
        self.codex_model = codex_model
        self.claude_model = claude_model

    def run_prover(self, task: StatementTask) -> ProverResult:
        prompt = self._build_prover_prompt(task)
        if self.prover_backend == "claude":
            payload = self._run_claude(
                prompt,
                PROVER_SCHEMA,
                role_name="prover",
                cwd=self.run_dir,
            )
        else:
            payload = self._run_codex(prompt, PROVER_SCHEMA, sandbox="workspace-write")
        return ProverResult(**payload)

    def run_verifier(
        self,
        task: StatementTask,
        prover_result: ProverResult,
        compile_ok: bool,
        compile_output: str,
    ) -> VerifierResult:
        lean_text = ""
        if prover_result.lean_path:
            lean_file = self.run_dir / prover_result.lean_path
            if lean_file.exists():
                lean_text = lean_file.read_text(
                    encoding="utf-8",
                    errors="ignore",
                )[:24000]
        prompt = self._build_verifier_prompt(
            task,
            prover_result,
            compile_ok,
            compile_output,
            lean_text,
        )
        if self.verifier_backend == "claude":
            payload = self._run_claude(prompt, VERIFIER_SCHEMA, role_name="verifier")
        else:
            payload = self._run_codex(prompt, VERIFIER_SCHEMA, sandbox="read-only")
        return VerifierResult(**payload)

    def _run_codex(self, prompt: str, schema: dict[str, Any], sandbox: str) -> dict[str, Any]:
        self._ensure_cli_available("codex", "Codex")
        with tempfile.TemporaryDirectory(prefix="paper-badger-") as tmp_dir:
            tmp = Path(tmp_dir)
            schema_path = tmp / "schema.json"
            out_path = tmp / "out.json"
            schema_path.write_text(json.dumps(schema), encoding="utf-8")
            cmd = [
                "codex",
                "exec",
                "--skip-git-repo-check",
                "--cd",
                str(self.run_dir),
                "--sandbox",
                sandbox,
                "--output-schema",
                str(schema_path),
                "--output-last-message",
                str(out_path),
                "-",
            ]
            if sandbox == "workspace-write":
                cmd.insert(2, "--full-auto")
            if self.codex_model:
                cmd[2:2] = ["--model", self.codex_model]
            try:
                result = subprocess.run(
                    cmd,
                    input=prompt,
                    text=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    timeout=self.agent_timeout_seconds,
                    check=False,
                )
            except subprocess.TimeoutExpired as exc:
                timeout = self.agent_timeout_seconds
                raise AgentTimeoutError(
                    f"Codex agent timed out after {timeout}s.",
                ) from exc
            output = result.stdout or ""
            if result.returncode != 0:
                if self._looks_like_transport_error(output):
                    raise AgentTransportError(
                        self._transport_error_message("Codex", output),
                    )
                raise AgentInvocationError(
                    self._invocation_error_message("Codex", result.returncode, output),
                )
            try:
                result_data: dict[str, Any] = json.loads(out_path.read_text(encoding="utf-8"))
                return result_data
            except FileNotFoundError as exc:
                raise AgentInvocationError("Codex agent finished without writing structured output.") from exc
            except json.JSONDecodeError as exc:
                if self._looks_like_transport_error(output):
                    raise AgentTransportError(self._transport_error_message("Codex", output)) from exc
                raise AgentInvocationError("Codex agent returned invalid or empty structured output.") from exc

    def _run_claude(
        self,
        prompt: str,
        schema: dict[str, Any],
        role_name: str,
        cwd: Path | None = None,
    ) -> dict[str, Any]:
        self._ensure_cli_available("claude", "Claude")
        cmd = ["claude", "-p"]
        if self.claude_model:
            cmd.extend(["--model", self.claude_model])
        schema_text = json.dumps(schema)
        cmd.append(
            f"{prompt}\n\nReturn only JSON matching this schema:\n{schema_text}",
        )
        try:
            result = subprocess.run(
                cmd,
                text=True,
                capture_output=True,
                check=True,
                timeout=self.agent_timeout_seconds,
                cwd=cwd,
            )
        except subprocess.TimeoutExpired as exc:
            timeout = self.agent_timeout_seconds
            raise AgentTimeoutError(
                f"Claude {role_name} timed out after {timeout}s.",
            ) from exc
        except subprocess.CalledProcessError as exc:
            output = (exc.stdout or "") + ("\n" + exc.stderr if exc.stderr else "")
            if self._looks_like_transport_error(output):
                raise AgentTransportError(
                    self._transport_error_message("Claude", output),
                ) from exc
            raise AgentInvocationError(
                self._invocation_error_message(
                    "Claude",
                    exc.returncode,
                    output,
                ),
            ) from exc
        try:
            return self._extract_json_object(result.stdout)
        except json.JSONDecodeError as exc:
            if self._looks_like_transport_error(result.stdout or ""):
                raise AgentTransportError(
                    self._transport_error_message(
                        "Claude",
                        result.stdout or "",
                    ),
                ) from exc
            raise AgentInvocationError(f"Claude {role_name} returned invalid JSON.") from exc

    @staticmethod
    def _looks_like_transport_error(output: str) -> bool:
        haystack = output.lower()
        needles = [
            "stream disconnected",
            "error sending request",
            "reconnecting...",
            "connection reset",
            "connection refused",
            "temporarily unavailable",
            "tls handshake",
            "timed out waiting for headers",
            "backend-api/codex/responses",
            "backend-api/codex/models",
        ]
        return any(needle in haystack for needle in needles)

    @staticmethod
    def _ensure_cli_available(command: str, label: str) -> None:
        if shutil.which(command):
            return
        raise AgentInvocationError(f"{label} CLI not found on PATH.")

    @staticmethod
    def _transport_error_message(agent_name: str, output: str) -> str:
        tail = AgentRunner._tail_lines(output)
        suffix = f" Details: {tail}" if tail else ""
        return f"{agent_name} transport disconnected before returning a result.{suffix}"

    @staticmethod
    def _invocation_error_message(agent_name: str, returncode: int, output: str) -> str:
        tail = AgentRunner._tail_lines(output)
        suffix = f" Details: {tail}" if tail else ""
        return f"{agent_name} agent failed with exit status {returncode}.{suffix}"

    @staticmethod
    def _tail_lines(output: str, max_lines: int = 4, max_chars: int = 500) -> str:
        lines = [line.strip() for line in output.splitlines() if line.strip()]
        if not lines:
            return ""
        joined = " | ".join(lines[-max_lines:])
        if len(joined) > max_chars:
            return joined[-max_chars:]
        return joined

    @staticmethod
    def _extract_json_object(output: str) -> dict[str, Any]:
        decoder = json.JSONDecoder()
        last_dict: dict[str, Any] | None = None
        for start, char in enumerate(output):
            if char != "{":
                continue
            try:
                parsed: Any = decoder.raw_decode(output[start:])
                value: Any = parsed[0]
                end: int = parsed[1]
            except json.JSONDecodeError:
                continue
            if isinstance(value, dict):
                trailing = output[start + end :].strip()
                if not trailing:
                    return value  # type: ignore[no-any-return,unused-ignore]
                last_dict = value  # type: ignore[assignment,unused-ignore]
        if last_dict is not None:
            return last_dict
        raise json.JSONDecodeError(
            "No top-level JSON object found.",
            output,
            0,
        )

    def _build_prover_prompt(self, task: StatementTask) -> str:
        root = self.root_module or "Paper"
        stem = sanitize_label_stem(task.label)
        kind_dir = task.kind.capitalize()
        lean_target = f"{root}/Formalizations/{kind_dir}/{stem}.lean"
        paper_context = statement_context_summary(
            self.paper_dir,
            task.tex_path,
            task.sequence_index,
        )
        history = "\n".join(f"- {entry}" for entry in task.attempt_history) or "- none"
        retry_context = ""
        if task.summary or task.last_error or task.lean_path:
            retry_context = f"""

Previous attempt context:
- previous lean file: {task.lean_path or ""}
- verifier / supervisor feedback: {task.last_error or task.summary or ""}
- recent attempt history:
{history}

Address that feedback directly in this attempt. If the formalization is mathematically correct but the verifier lacked context, make that context explicit in your final summary.
""".rstrip()
        return f"""
You are the prover in a paper-badger formalization workflow.

Work only on this single statement until one of the following is true:
1. the proof is formalized cleanly enough to be a verified candidate,
2. the statement is formalized exactly but the proof is still incomplete, or
3. further progress would genuinely take LLM-weeks.

Rules:
- Work in the current repository only.
- Do not edit files under `paper/`; the supervisor is the only component that inserts badges or changes the paper source.
- Keep `PROVER.md` updated with concise progress notes as you go.
- Keep `TODO.md` current.
- If the paper has a real ambiguity or mistake, append it to `ISSUES.md`.
- Create a human-readable Lean file for this statement.
- Do not leave `sorry`.
- Fix linter warnings in files you touch.
- Prefer creating or extending small supporting libraries rather than a monolithic file.
- When a theorem is not yet proved, it is acceptable to formalize the statement exactly in a non-verified way, but be explicit about that in the final JSON.

Statement metadata:
- kind: {task.kind}
- label: {task.label}
- title: {task.title or ""}
- tex file: {task.tex_path}
- suggested lean file: {lean_target}
{retry_context}

Statement excerpt:
{task.source_excerpt}

Bounded paper context:
{paper_context or "- none"}
""".strip()

    def _build_verifier_prompt(
        self,
        task: StatementTask,
        prover_result: ProverResult,
        compile_ok: bool,
        compile_output: str,
        lean_text: str,
    ) -> str:
        paper_context = statement_context_summary(
            self.paper_dir,
            task.tex_path,
            task.sequence_index,
        )
        history = "\n".join(f"- {entry}" for entry in task.attempt_history) or "- none"
        return f"""
You are the verifier in a paper-badger formalization workflow.

Decide whether this item deserves a `verified` badge, a `formalized` badge, should be retried, or should be classified as LLM-weeks-scale.

Important constraints:
- Use only the materials included in this prompt.
- Do not inspect the workspace.
- Do not run commands.
- Return a decision immediately from the provided excerpt, Lean text, and compile result.
- If the provided material is insufficient to justify `verified`, prefer `formalized` or `retry`.

Standards:
- The Lean statement must match the paper statement exactly in mathematical content.
- Ambient assumptions may be justified by the bounded paper context when that context explicitly states them.
- Reject additional hypotheses, weakened conclusions, or hidden assumptions.
- `verified` requires a complete proof and successful compilation.
- `formalized` means the statement is exact but the proof is not complete enough for `verified`.
- Assess globally: use the recent attempt history to judge whether the task is still improving or whether the remaining gap honestly looks like LLM-weeks.
- Return `weeks_scale` only when that global assessment justifies it, not merely because there have been several attempts.

Paper statement metadata:
- kind: {task.kind}
- label: {task.label}
- title: {task.title or ""}
- tex file: {task.tex_path}

Bounded paper context:
{paper_context or "- none"}

Recent attempt history:
{history}

Paper excerpt:
{task.source_excerpt}

Prover summary:
{prover_result.summary}

Compile succeeded: {compile_ok}
Compile output:
{compile_output[:2000]}

Lean file contents:
{lean_text[:12000]}
""".strip()



from __future__ import annotations

import gzip
import io
import shutil
import tarfile
import urllib.error
import urllib.request
from pathlib import Path  # noqa: TC003

ARXIV_SOURCE_URLS = (
    "https://export.arxiv.org/e-print/{arxiv_id}",
    "https://arxiv.org/e-print/{arxiv_id}",
    "https://export.arxiv.org/src/{arxiv_id}",
    "https://arxiv.org/src/{arxiv_id}",
)


def download_and_extract_arxiv_source(arxiv_id: str, output_dir: Path) -> Path:
    output_dir.mkdir(parents=True, exist_ok=True)
    payload = _download_source_payload(arxiv_id)
    _extract_payload(payload, output_dir)
    return output_dir


def _download_source_payload(arxiv_id: str) -> bytes:
    headers = {"User-Agent": "paper-badger/0.1 (+https://github.com/BioDisCo/paper-badger)"}
    errors: list[str] = []
    for template in ARXIV_SOURCE_URLS:
        url = template.format(arxiv_id=arxiv_id)
        req = urllib.request.Request(url, headers=headers)
        try:
            with urllib.request.urlopen(req, timeout=60) as response:
                payload: bytes = response.read()
                if not payload:
                    errors.append(f"{url}: empty response")
                    continue
                if _looks_like_html(payload):
                    errors.append(f"{url}: returned HTML instead of source")
                    continue
                return payload
        except urllib.error.URLError as exc:
            errors.append(f"{url}: {exc}")
    raise RuntimeError("failed to download arXiv source:\n" + "\n".join(errors))


def _extract_payload(payload: bytes, output_dir: Path) -> None:
    if _is_tar_archive(payload):
        with tarfile.open(fileobj=io.BytesIO(payload), mode="r:*") as archive:
            _safe_extract_tar(archive, output_dir)
        return
    if _is_gzip_stream(payload):
        decompressed = gzip.decompress(payload)
        if _is_tar_archive(decompressed):
            with tarfile.open(fileobj=io.BytesIO(decompressed), mode="r:*") as archive:
                _safe_extract_tar(archive, output_dir)
            return
        (output_dir / "main.tex").write_bytes(decompressed)
        return
    (output_dir / "main.tex").write_bytes(payload)


def _looks_like_html(payload: bytes) -> bool:
    head = payload[:512].lower()
    return b"<html" in head or b"<!doctype html" in head


def _is_tar_archive(payload: bytes) -> bool:
    try:
        with tarfile.open(fileobj=io.BytesIO(payload), mode="r:*"):
            return True
    except tarfile.TarError:
        return False


def _is_gzip_stream(payload: bytes) -> bool:
    return payload[:2] == b"\x1f\x8b"


def _safe_extract_tar(archive: tarfile.TarFile, output_dir: Path) -> None:
    output_dir = output_dir.resolve()
    for member in archive.getmembers():
        if member.issym() or member.islnk():
            raise RuntimeError(f"unsafe archive member (symlink): {member.name}")
        member_path = (output_dir / member.name).resolve()
        if not member_path.is_relative_to(output_dir):
            raise RuntimeError(f"unsafe archive member: {member.name}")
    archive.extractall(output_dir)


def wipe_directory(path: Path) -> None:
    if path.exists():
        shutil.rmtree(path)



import json
from pathlib import Path
from unittest.mock import patch

from paper_badger.agents import AgentTransportError, ProverResult, VerifierResult
from paper_badger.models import RunState, StatementTask
from paper_badger.workflow import (
    _append_attempt_history,
    _badge_target_for_task,
    _load_or_initialize_state,
    run_formalization,
)


def _task(
    label: str = "thm:main",
    kind: str = "theorem",
    status: str = "pending",
    sequence_index: int = 0,
) -> StatementTask:
    return StatementTask(
        task_id=f"main.tex::{sequence_index}::{label}",
        tex_path="main.tex",
        env_name=kind,
        label=label,
        title="Main",
        sequence_index=sequence_index,
        kind=kind,
        source_excerpt="If m and n are even, then m+n is even.",
        status=status,
    )


def _write_state(run_dir: Path, task: StatementTask) -> None:
    paper_dir = run_dir / "paper"
    paper_dir.mkdir(parents=True, exist_ok=True)
    (paper_dir / "main.tex").write_text(
        r"""
\documentclass{article}
\begin{document}
\begin{theorem}[Main]
\label{thm:main}
If $m$ and $n$ are even, then $m+n$ is even.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )
    root = run_dir / "Paper"
    root.mkdir(parents=True, exist_ok=True)
    (root / "Basic.lean").write_text("namespace Paper\nend Paper\n", encoding="utf-8")
    state = RunState(
        arxiv_id="paper-id",
        run_dir=str(run_dir),
        paper_dir=str(paper_dir),
        main_tex=str(paper_dir / "main.tex"),
        repo_url="https://github.com/example/repo",
        branch="main",
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="auto",
        explicit_repo_target=False,
        is_local_paper=False,
        root_module="Paper",
        tasks=[task],
    )
    state.state_path.write_text(json.dumps(state.to_dict(), indent=2, sort_keys=True), encoding="utf-8")


def test_attempt_history_keeps_recent_entries() -> None:
    task = _task()

    for index in range(15):
        _append_attempt_history(task, f"Attempt {index + 1}: note {index + 1}")

    assert len(task.attempt_history) == 12
    assert task.attempt_history[0] == "Attempt 4: note 4"
    assert task.attempt_history[-1] == "Attempt 15: note 15"


def test_load_existing_state_resumes_interrupted_proving_task(tmp_path: Path) -> None:
    run_dir = tmp_path / "paper-id"
    task = _task(status="proving")
    _write_state(run_dir, task)

    state = _load_or_initialize_state(
        arxiv_id="paper-id",
        run_dir=run_dir,
        local_paper_dir=None,
        repo_url=None,
        branch=None,
        prover_backend="codex",
        verifier_backend="claude",
        badge_link_mode="auto",
        verified_badges_repo=None,
    )

    assert state.verifier_backend == "claude"
    assert state.tasks[0].status == "retry"
    assert "interrupted proving task" in state.tasks[0].attempt_history[-1]
    assert state.tasks[0].summary == "Resuming after interrupted proving attempt."


def test_load_existing_state_removes_stale_badge_for_nonfinal_task(tmp_path: Path) -> None:
    run_dir = tmp_path / "paper-id"
    task = _task(status="proving")
    _write_state(run_dir, task)
    main_tex = run_dir / "paper" / "main.tex"
    main_tex.write_text(
        r"""
\documentclass{article}
\begin{document}
\begin{theorem}[Main \leanproof{Paper/Formalizations/Theorem/ThmMain.lean}]
\label{thm:main}
If $m$ and $n$ are even, then $m+n$ is even.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    state = _load_or_initialize_state(
        arxiv_id="paper-id",
        run_dir=run_dir,
        local_paper_dir=None,
        repo_url=None,
        branch=None,
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="auto",
        verified_badges_repo=None,
    )

    updated = main_tex.read_text(encoding="utf-8")

    assert state.tasks[0].status == "retry"
    assert r"\leanproof{Paper/Formalizations/Theorem/ThmMain.lean}" not in updated
    assert r"\begin{theorem}[Main]" in updated


def test_local_paper_initialization_extracts_tasks_and_inserts_badges_package(tmp_path: Path) -> None:
    run_dir = tmp_path / "local-paper"
    local_paper_dir = tmp_path / "sample"
    local_paper_dir.mkdir()
    main_tex = local_paper_dir / "main.tex"
    main_tex.write_text(
        r"""
\documentclass{article}
\title{A Note About Naturals}
\begin{document}
\begin{definition}[Even]
\label{def:even}
A natural number $n$ is even if $n = 2k$ for some natural number $k$.
\end{definition}
\begin{theorem}[Main]
\label{thm:main}
If $n$ is even, then $n + 2$ is even.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )
    badges_repo = tmp_path / "badges"
    badges_repo.mkdir()

    def fake_init(project_dir: Path) -> None:
        (project_dir / "lakefile.toml").write_text('name = "local_paper"\n', encoding="utf-8")
        root = project_dir / "LocalPaper"
        root.mkdir(parents=True, exist_ok=True)
        (root / "Basic.lean").write_text("namespace LocalPaper\nend LocalPaper\n", encoding="utf-8")
        for subdir in ["Definition", "Notation", "Lemma", "Proposition", "Corollary", "Theorem"]:
            (root / "Formalizations" / subdir).mkdir(parents=True, exist_ok=True)

    with (
        patch("paper_badger.workflow._initialize_lean_project", side_effect=fake_init),
        patch("paper_badger.workflow.ensure_verified_badges_repo", return_value=badges_repo),
        patch("paper_badger.workflow.copy_verified_badges_assets"),
        patch(
            "paper_badger.workflow.infer_repo_url_and_branch",
            return_value=("https://github.com/example/repo", "main"),
        ),
    ):
        state = _load_or_initialize_state(
            arxiv_id="local-paper",
            run_dir=run_dir,
            local_paper_dir=local_paper_dir,
            repo_url=None,
            branch=None,
            prover_backend="codex",
            verifier_backend="codex",
            badge_link_mode="auto",
            verified_badges_repo=badges_repo,
        )

    assert state.root_module == "LocalPaper"
    assert state.is_local_paper
    assert [task.label for task in state.tasks] == ["def:even", "thm:main"]
    updated_main = (run_dir / "paper" / "main.tex").read_text(encoding="utf-8")
    assert r"\usepackage{verified-badges}" in updated_main


def test_run_formalization_retries_then_verifies_and_inserts_badge(tmp_path: Path) -> None:
    base_dir = tmp_path
    run_dir = base_dir / "paper-id"
    task = _task()
    _write_state(run_dir, task)
    theorem_dir = run_dir / "Paper" / "Formalizations" / "Theorem"
    theorem_dir.mkdir(parents=True, exist_ok=True)
    (theorem_dir / "ThmMain.lean").write_text("namespace Paper\nend Paper\n", encoding="utf-8")

    class FakeRunner:
        verifier_calls = 0

        def __init__(self, **_: object) -> None:
            pass

        def run_prover(self, task: StatementTask) -> ProverResult:
            return ProverResult(
                status="verified_candidate",
                summary="Built theorem.",
                lean_path="Paper/Formalizations/Theorem/ThmMain.lean",
                proof_complete=True,
                weeks_scale=False,
            )

        def run_verifier(
            self,
            task: StatementTask,
            prover_result: ProverResult,
            compile_ok: bool,
            compile_output: str,
        ) -> VerifierResult:
            FakeRunner.verifier_calls += 1
            if FakeRunner.verifier_calls == 1:
                return VerifierResult(
                    decision="retry",
                    summary="Need one more exact-match clarification.",
                    exact_match=False,
                    extra_hypotheses=True,
                )
            return VerifierResult(
                decision="verified",
                summary="Exact and complete.",
                exact_match=True,
                extra_hypotheses=False,
            )

    with (
        patch("paper_badger.workflow.AgentRunner", FakeRunner),
        patch(
            "paper_badger.workflow._compile_candidate",
            return_value=(True, "ok"),
        ),
        patch(
            "paper_badger.workflow.build_blob_url",
            return_value="https://github.com/example/repo/blob/main/Paper/Formalizations/Theorem/ThmMain.lean",
        ),
    ):
        state = run_formalization(
            arxiv_id="paper-id",
            base_dir=base_dir,
            local_paper_dir=None,
            repo_url="https://github.com/example/repo",
            branch="main",
            prover_backend="codex",
            verifier_backend="codex",
            badge_link_mode="auto",
            verified_badges_repo=None,
            agent_timeout_seconds=10,
            codex_model=None,
            claude_model=None,
            max_attempts_per_task=5,
            max_tasks=None,
        )

    task = state.tasks[0]
    updated_main = (run_dir / "paper" / "main.tex").read_text(encoding="utf-8")

    assert task.status == "verified"
    assert task.attempts == 2
    assert "retry requested" in task.attempt_history[0]
    assert "verifier accepted verified" in task.attempt_history[-1]
    assert (
        r"\leanproof{https://github.com/example/repo/blob/main/Paper/Formalizations/Theorem/ThmMain.lean}"
        in updated_main
    )


def test_transport_failure_does_not_consume_substantive_attempt(tmp_path: Path) -> None:
    base_dir = tmp_path
    run_dir = base_dir / "paper-id"
    task = _task()
    _write_state(run_dir, task)
    theorem_dir = run_dir / "Paper" / "Formalizations" / "Theorem"
    theorem_dir.mkdir(parents=True, exist_ok=True)
    (theorem_dir / "ThmMain.lean").write_text("namespace Paper\nend Paper\n", encoding="utf-8")

    class FakeRunner:
        prover_calls = 0

        def __init__(self, **_: object) -> None:
            pass

        def run_prover(self, task: StatementTask) -> ProverResult:
            FakeRunner.prover_calls += 1
            if FakeRunner.prover_calls == 1:
                raise AgentTransportError("Codex transport disconnected before returning a result.")
            return ProverResult(
                status="verified_candidate",
                summary="Built theorem.",
                lean_path="Paper/Formalizations/Theorem/ThmMain.lean",
                proof_complete=True,
                weeks_scale=False,
            )

        def run_verifier(
            self,
            task: StatementTask,
            prover_result: ProverResult,
            compile_ok: bool,
            compile_output: str,
        ) -> VerifierResult:
            return VerifierResult(
                decision="verified",
                summary="Exact and complete.",
                exact_match=True,
                extra_hypotheses=False,
            )

    with (
        patch("paper_badger.workflow.AgentRunner", FakeRunner),
        patch(
            "paper_badger.workflow._compile_candidate",
            return_value=(True, "ok"),
        ),
        patch(
            "paper_badger.workflow.build_blob_url",
            return_value="https://github.com/example/repo/blob/main/Paper/Formalizations/Theorem/ThmMain.lean",
        ),
        patch("paper_badger.workflow._sleep_after_transport_failure"),
    ):
        state = run_formalization(
            arxiv_id="paper-id",
            base_dir=base_dir,
            local_paper_dir=None,
            repo_url="https://github.com/example/repo",
            branch="main",
            prover_backend="codex",
            verifier_backend="codex",
            badge_link_mode="auto",
            verified_badges_repo=None,
            agent_timeout_seconds=10,
            codex_model=None,
            claude_model=None,
            max_attempts_per_task=5,
            max_tasks=None,
        )

    task = state.tasks[0]

    assert task.status == "verified"
    assert task.attempts == 1
    assert task.transport_failures == 0
    assert "Transport failure while starting attempt 1" in task.attempt_history[0]


def test_missing_prover_output_file_retries_before_verifier(tmp_path: Path) -> None:
    base_dir = tmp_path
    run_dir = base_dir / "paper-id"
    task = _task()
    _write_state(run_dir, task)

    class FakeRunner:
        verifier_calls = 0

        def __init__(self, **_: object) -> None:
            pass

        def run_prover(self, task: StatementTask) -> ProverResult:
            return ProverResult(
                status="verified_candidate",
                summary="Wrote the theorem file.",
                lean_path="Paper/Formalizations/Theorem/Missing.lean",
                proof_complete=True,
                weeks_scale=False,
            )

        def run_verifier(
            self,
            task: StatementTask,
            prover_result: ProverResult,
            compile_ok: bool,
            compile_output: str,
        ) -> VerifierResult:
            FakeRunner.verifier_calls += 1
            return VerifierResult(
                decision="verified",
                summary="Should never be called.",
                exact_match=True,
                extra_hypotheses=False,
            )

    with patch("paper_badger.workflow.AgentRunner", FakeRunner):
        state = run_formalization(
            arxiv_id="paper-id",
            base_dir=base_dir,
            local_paper_dir=None,
            repo_url="https://github.com/example/repo",
            branch="main",
            prover_backend="codex",
            verifier_backend="codex",
            badge_link_mode="auto",
            verified_badges_repo=None,
            agent_timeout_seconds=10,
            codex_model=None,
            claude_model=None,
            max_attempts_per_task=1,
            max_tasks=None,
        )

    task = state.tasks[0]

    assert FakeRunner.verifier_calls == 0
    assert task.status == "retry"
    assert task.attempts == 1
    assert task.summary == "Lean file does not exist: Paper/Formalizations/Theorem/Missing.lean"
    assert "retry requested before verification" in task.attempt_history[-1]


def test_badge_target_uses_local_relative_path_for_local_runs(tmp_path: Path) -> None:
    run_dir = tmp_path / "run"
    paper_dir = run_dir / "paper"
    paper_dir.mkdir(parents=True)
    state = RunState(
        arxiv_id="local-paper",
        run_dir=str(run_dir),
        paper_dir=str(paper_dir),
        main_tex=str(paper_dir / "main.tex"),
        repo_url="https://github.com/example/repo",
        branch="main",
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="auto",
        explicit_repo_target=False,
        is_local_paper=True,
        root_module="Paper",
        tasks=[],
    )

    target = _badge_target_for_task(state, "Paper/Formalizations/Theorem/ThmMain.lean")

    assert target == (run_dir / "Paper" / "Formalizations" / "Theorem" / "ThmMain.lean").resolve().as_uri()


def test_badge_target_uses_local_file_uri_for_non_explicit_auto_mode(tmp_path: Path) -> None:
    run_dir = tmp_path / "run"
    paper_dir = run_dir / "paper"
    paper_dir.mkdir(parents=True)
    state = RunState(
        arxiv_id="arxiv-paper",
        run_dir=str(run_dir),
        paper_dir=str(paper_dir),
        main_tex=str(paper_dir / "main.tex"),
        repo_url="https://github.com/example/repo",
        branch="main",
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="auto",
        explicit_repo_target=False,
        is_local_paper=False,
        root_module="Paper",
        tasks=[],
    )

    target = _badge_target_for_task(state, "Paper/Formalizations/Theorem/ThmMain.lean")

    assert target == (run_dir / "Paper" / "Formalizations" / "Theorem" / "ThmMain.lean").resolve().as_uri()


def test_badge_target_uses_github_for_explicit_auto_mode(tmp_path: Path) -> None:
    run_dir = tmp_path / "run"
    paper_dir = run_dir / "paper"
    paper_dir.mkdir(parents=True)
    state = RunState(
        arxiv_id="arxiv-paper",
        run_dir=str(run_dir),
        paper_dir=str(paper_dir),
        main_tex=str(paper_dir / "main.tex"),
        repo_url="https://github.com/example/repo",
        branch="main",
        prover_backend="codex",
        verifier_backend="codex",
        badge_link_mode="auto",
        explicit_repo_target=True,
        is_local_paper=False,
        root_module="Paper",
        tasks=[],
    )

    with patch(
        "paper_badger.workflow.build_blob_url",
        return_value="https://github.com/example/repo/blob/main/Paper/Formalizations/Theorem/ThmMain.lean",
    ):
        target = _badge_target_for_task(state, "Paper/Formalizations/Theorem/ThmMain.lean")

    assert target == "https://github.com/example/repo/blob/main/Paper/Formalizations/Theorem/ThmMain.lean"



from __future__ import annotations

import logging
import re
from collections.abc import Iterator  # noqa: TC003
from dataclasses import dataclass
from pathlib import Path

logger = logging.getLogger(__name__)


ENV_KIND_MAP = {
    "definition": "definition",
    "def": "definition",
    "notation": "notation",
    "lemma": "lemma",
    "lem": "lemma",
    "proposition": "proposition",
    "prop": "proposition",
    "corollary": "corollary",
    "cor": "corollary",
    "theorem": "theorem",
    "thm": "theorem",
}

THEOREM_PRIORITY = {
    "definition": 0,
    "notation": 1,
    "lemma": 10,
    "proposition": 11,
    "corollary": 12,
    "theorem": 13,
}

BEGIN_RE = re.compile(r"\\begin\{(?P<env>[A-Za-z*]+)\}(?P<opt>\[[^\]]*\])?", re.MULTILINE)
LABEL_RE = re.compile(r"\\label\{(?P<label>[^}]+)\}")
BADGE_RE = re.compile(r"\\lean(?:proof|formalized)\{[^}]+\}")
TITLE_RE = re.compile(r"\\title\{(?P<title>[^}]*)\}")


@dataclass
class ExtractedStatement:
    task_id: str
    tex_path: str
    env_name: str
    label: str
    title: str | None
    sequence_index: int
    kind: str
    source_excerpt: str


@dataclass
class LocatedStatement:
    extracted: ExtractedStatement
    begin_start: int
    begin_end: int
    begin_text: str


def detect_main_tex(tex_files: list[Path]) -> Path:
    document_candidates: list[tuple[int, Path]] = []
    for path in tex_files:
        text = path.read_text(encoding="utf-8", errors="ignore")
        if "\\documentclass" in text and "\\begin{document}" in text:
            document_candidates.append((len(text), path))
    if document_candidates:
        return max(document_candidates, key=lambda item: item[0])[1]
    if not tex_files:
        raise RuntimeError("no .tex files found in paper source")
    return max(tex_files, key=lambda path: path.stat().st_size)


def find_tex_files(root: Path) -> list[Path]:
    return sorted(path for path in root.rglob("*.tex") if path.is_file())


def extract_statements_from_tree(paper_dir: Path) -> list[ExtractedStatement]:
    extracted: list[ExtractedStatement] = []
    for tex_path in find_tex_files(paper_dir):
        text = tex_path.read_text(encoding="utf-8", errors="ignore")
        text = _document_body(text)
        extracted.extend(_extract_statements_from_text(text, tex_path.relative_to(paper_dir)))
    extracted.sort(key=_task_sort_key)
    return extracted


def locate_statement(
    paper_dir: Path,
    task_id: str,
    tex_path: str,
    sequence_index: int,  # noqa: ARG001
) -> LocatedStatement:
    path = paper_dir / tex_path
    full_text = path.read_text(encoding="utf-8", errors="ignore")
    text, body_offset = _document_body_with_offset(full_text)
    statements = list(_iter_statement_matches(text, Path(tex_path), body_offset=body_offset))
    for match in statements:
        if match.extracted.task_id == task_id:
            return match
    raise RuntimeError(f"statement identity drifted for {task_id}")


def ensure_verified_badges_package(main_tex_path: Path) -> bool:
    text = main_tex_path.read_text(encoding="utf-8", errors="ignore")
    if "\\usepackage{verified-badges}" in text:
        return False
    marker = "\\begin{document}"
    idx = text.find(marker)
    if idx == -1:
        raise RuntimeError(f"failed to find {marker} in {main_tex_path}")
    insertion = "\\usepackage{verified-badges}\n"
    text = text[:idx] + insertion + text[idx:]
    main_tex_path.write_text(text, encoding="utf-8")
    return True


def insert_badge_for_task(paper_dir: Path, task_id: str, tex_path: str, sequence_index: int, badge_macro: str) -> bool:
    return set_badge_for_task(paper_dir, task_id, tex_path, sequence_index, badge_macro)


def set_badge_for_task(
    paper_dir: Path,
    task_id: str,
    tex_path: str,
    sequence_index: int,
    badge_macro: str | None,
) -> bool:
    location = locate_statement(paper_dir, task_id, tex_path, sequence_index)
    new_begin = _set_badge_macro(location.begin_text, badge_macro)
    if new_begin == location.begin_text:
        return False
    path = paper_dir / tex_path
    text = path.read_text(encoding="utf-8", errors="ignore")
    text = text[: location.begin_start] + new_begin + text[location.begin_end :]
    path.write_text(text, encoding="utf-8")
    return True


def statement_context_summary(paper_dir: Path, tex_path: str, sequence_index: int, max_previous: int = 2) -> str:
    path = paper_dir / tex_path
    full_text = path.read_text(encoding="utf-8", errors="ignore")
    body, _ = _document_body_with_offset(full_text)
    statements = list(_iter_statement_matches(body, Path(tex_path)))
    position = next(
        (index for index, item in enumerate(statements) if item.extracted.sequence_index == sequence_index), None
    )
    if position is None:
        return ""
    pieces: list[str] = []
    title_match = TITLE_RE.search(full_text)
    if title_match:
        title = _shorten_whitespace(title_match.group("title"))
        if title:
            pieces.append(f"Paper title: {title}")
    start_index = max(0, position - max_previous)
    previous = statements[start_index:position]
    if previous:
        pieces.append("Earlier statements in the same file:")
        for item in previous:
            extracted = item.extracted
            heading = f"{extracted.kind} `{extracted.label}`"
            if extracted.title:
                heading += f" ({extracted.title})"
            pieces.append(f"- {heading}: {extracted.source_excerpt}")
    return "\n".join(pieces)


def _extract_statements_from_text(text: str, tex_path: Path) -> list[ExtractedStatement]:
    return [location.extracted for location in _iter_statement_matches(text, tex_path)]


def _iter_statement_matches(text: str, tex_path: Path, body_offset: int = 0) -> Iterator[LocatedStatement]:
    sequence_index = 0
    for match in BEGIN_RE.finditer(text):
        raw_env = match.group("env")
        env_name = raw_env.rstrip("*").lower()
        kind = ENV_KIND_MAP.get(env_name)
        if kind is None:
            continue
        end_tag = f"\\end{{{raw_env}}}"
        end_idx = text.find(end_tag, match.end())
        if end_idx == -1:
            continue
        body = text[match.end() : end_idx]
        body_without_comments = _strip_comments(body)
        title = _normalize_title(match.group("opt"))
        label_match = LABEL_RE.search(body)
        label = label_match.group("label") if label_match else _synthetic_label(env_name, title, sequence_index)
        excerpt = _shorten_whitespace(body_without_comments)
        excerpt_without_label = _shorten_whitespace(LABEL_RE.sub("", body_without_comments))
        if not excerpt_without_label:
            sequence_index += 1
            continue
        task_id = f"{tex_path.as_posix()}::{sequence_index}::{label}"
        yield LocatedStatement(
            extracted=ExtractedStatement(
                task_id=task_id,
                tex_path=tex_path.as_posix(),
                env_name=raw_env,
                label=label,
                title=title,
                sequence_index=sequence_index,
                kind=kind,
                source_excerpt=excerpt[:1600],
            ),
            begin_start=body_offset + match.start(),
            begin_end=body_offset + match.end(),
            begin_text=match.group(0),
        )
        sequence_index += 1


def _normalize_title(optional_arg: str | None) -> str | None:
    if optional_arg is None:
        return None
    stripped = optional_arg.strip()[1:-1].strip()
    stripped = BADGE_RE.sub("", stripped).strip()
    return stripped or None


def _synthetic_label(env_name: str, title: str | None, sequence_index: int) -> str:
    stem = _slugify(title or f"{env_name}-{sequence_index}")
    return f"synthetic:{env_name}:{stem}"


def _slugify(text: str) -> str:
    lowered = text.lower()
    lowered = re.sub(r"\\[A-Za-z]+", "", lowered)
    lowered = re.sub(r"[^a-z0-9]+", "-", lowered)
    lowered = lowered.strip("-")
    return lowered or "item"


def sanitize_label_stem(label: str) -> str:
    return _slugify(label.replace(":", "-"))


def _task_sort_key(item: ExtractedStatement) -> tuple[int, int, str, int]:
    synthetic_penalty = 1 if item.label.startswith("synthetic:") else 0
    return (synthetic_penalty, THEOREM_PRIORITY.get(item.kind, 100), item.tex_path, item.sequence_index)


def _strip_comments(text: str) -> str:
    return re.sub(r"(?<!\\)%[^\n]*", "", text)


def _shorten_whitespace(text: str) -> str:
    return re.sub(r"\s+", " ", text).strip()


def _set_badge_macro(begin_text: str, badge_macro: str | None) -> str:
    optional_match = re.search(r"\[(?P<content>[^\]]*)\]$", begin_text)
    if optional_match is None:
        if badge_macro is None:
            return begin_text
        return begin_text + f"[{badge_macro}]"

    prefix = begin_text[: optional_match.start()]
    content = optional_match.group("content")
    content = BADGE_RE.sub("", content)
    content = re.sub(r"\s+", " ", content).strip()
    if badge_macro is not None:
        content = f"{content} {badge_macro}".strip()
    if not content:
        return prefix
    return f"{prefix}[{content}]"


def _document_body(text: str) -> str:
    return _document_body_with_offset(text)[0]


def _document_body_with_offset(text: str) -> tuple[str, int]:
    start = text.find(r"\begin{document}")
    if start == -1:
        logger.warning("no \\begin{document} found; scanning entire file for statements")
        return text, 0
    start += len(r"\begin{document}")
    end = text.find(r"\end{document}", start)
    if end == -1:
        end = len(text)
    return text[start:end], start



from __future__ import annotations

import argparse
import sys
from pathlib import Path

from .dashboard import load_state, monitor_run, render_dashboard
from .workflow import run_formalization


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Formalize arXiv proofs and insert verification badges.",
    )
    subparsers = parser.add_subparsers(dest="command")

    run_parser = subparsers.add_parser(
        "run",
        help="Start or resume a formalization run",
    )
    run_parser.add_argument(
        "paper_id",
        help="Run identifier, e.g. 2401.01234 or trivial-even",
    )
    run_parser.add_argument(
        "--runs-dir",
        default="runs",
        help="Directory for per-paper workspaces",
    )
    run_parser.add_argument(
        "--paper-dir",
        default=None,
        help=(
            "Local directory containing LaTeX paper sources; "
            "if omitted, the positional identifier is treated as an arXiv ID"
        ),
    )
    run_parser.add_argument(
        "--repo-url",
        default=None,
        help="GitHub repository URL used for badge links",
    )
    run_parser.add_argument(
        "--branch",
        default=None,
        help="Git branch name used for badge links",
    )
    run_parser.add_argument(
        "--badge-link-mode",
        choices=["auto", "local", "github"],
        default="local",
        help="How badge links should be generated",
    )
    run_parser.add_argument(
        "--prover-backend",
        choices=["claude", "codex"],
        default="codex",
        help="Prover backend",
    )
    run_parser.add_argument(
        "--verifier-backend",
        choices=["claude", "codex"],
        default="claude",
        help="Verifier backend",
    )
    run_parser.add_argument(
        "--verified-badges-repo",
        default=None,
        help="Path to a local verified-badges checkout",
    )
    run_parser.add_argument(
        "--agent-timeout-seconds",
        type=int,
        default=1800,
        help="Max seconds per prover/verifier invocation",
    )
    run_parser.add_argument(
        "--codex-model",
        default=None,
        help="Optional Codex model override",
    )
    run_parser.add_argument(
        "--claude-model",
        default=None,
        help="Optional Claude model override",
    )
    run_parser.add_argument(
        "--max-attempts-per-task",
        type=int,
        default=None,
        help="Per-task attempt cap; unresolved tasks stay in retry",
    )
    run_parser.add_argument(
        "--max-tasks",
        type=int,
        default=None,
        help="Process at most this many tasks in one invocation",
    )

    monitor_parser = subparsers.add_parser(
        "monitor",
        help="Render a live dashboard for an existing run",
    )
    monitor_parser.add_argument(
        "paper_id",
        help="Run identifier to inspect",
    )
    monitor_parser.add_argument(
        "--runs-dir",
        default="runs",
        help="Directory containing per-paper workspaces",
    )
    monitor_parser.add_argument(
        "--interval-seconds",
        type=float,
        default=5.0,
        help="Refresh interval for the live dashboard",
    )
    monitor_parser.add_argument(
        "--once",
        action="store_true",
        help="Render a single snapshot and exit",
    )
    return parser


def main() -> None:
    parser = build_parser()
    argv = sys.argv[1:]
    if argv and argv[0] not in {"run", "monitor", "-h", "--help"}:
        if argv[0].startswith("-"):
            parser.parse_args(argv)
            return
        argv = ["run", *argv]
    args = parser.parse_args(argv)
    if args.command == "monitor":
        run_dir = Path(args.runs_dir).resolve() / args.paper_id
        if args.once:
            state = load_state(run_dir)
            state.run_dir = str(run_dir)
            print(render_dashboard(state))
            return
        monitor_run(run_dir, args.interval_seconds)
        return
    state = run_formalization(
        arxiv_id=args.paper_id,
        base_dir=Path(args.runs_dir).resolve(),
        local_paper_dir=(Path(args.paper_dir).resolve() if args.paper_dir else None),
        repo_url=args.repo_url,
        branch=args.branch,
        prover_backend=args.prover_backend,
        verifier_backend=args.verifier_backend,
        badge_link_mode=args.badge_link_mode,
        verified_badges_repo=(Path(args.verified_badges_repo).resolve() if args.verified_badges_repo else None),
        agent_timeout_seconds=args.agent_timeout_seconds,
        codex_model=args.codex_model,
        claude_model=args.claude_model,
        max_attempts_per_task=args.max_attempts_per_task,
        max_tasks=args.max_tasks,
    )
    completed = sum(1 for task in state.tasks if task.status in {"verified", "formalized"})
    weeks = sum(1 for task in state.tasks if task.status == "weeks_scale")
    retries = sum(1 for task in state.tasks if task.status == "retry")
    pending = sum(1 for task in state.tasks if task.status in {"pending", "proving"})
    print(
        f"Run completed for {state.arxiv_id}: "
        f"{completed} completed, {retries} retry, "
        f"{pending} pending, {weeks} weeks-scale."
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import logging
import time
from pathlib import Path

from .models import RunState, StatementTask

logger = logging.getLogger(__name__)


def load_state(run_dir: Path) -> RunState:
    state_path = run_dir / "state.json"
    payload = json.loads(state_path.read_text(encoding="utf-8"))
    return RunState.from_dict(payload)


def render_dashboard(state: RunState) -> str:
    tasks = state.tasks
    completed = sum(1 for task in tasks if task.status in {"verified", "formalized"})
    verified = sum(1 for task in tasks if task.status == "verified")
    formalized = sum(1 for task in tasks if task.status == "formalized")
    retry = sum(1 for task in tasks if task.status == "retry")
    proving = sum(1 for task in tasks if task.status == "proving")
    pending = sum(1 for task in tasks if task.status == "pending")
    weeks = sum(1 for task in tasks if task.status == "weeks_scale")
    current = next((task for task in tasks if task.status == "proving"), None)
    recent = [task for task in tasks if task.attempt_history]
    recent.sort(key=lambda task: task.attempts, reverse=True)

    lines = [
        f"Run: {state.arxiv_id}",
        f"Run dir: {state.run_dir}",
        (
            "Summary: "
            f"{completed}/{len(tasks)} completed "
            f"({verified} verified, {formalized} formalized), "
            f"{retry} retry, {proving} proving, {pending} pending, {weeks} weeks-scale"
        ),
        f"Backends: prover={state.prover_backend}, verifier={state.verifier_backend}, links={state.badge_link_mode}",
        "",
        "Current Task:",
    ]
    if current is None:
        lines.append("- none")
    else:
        lines.extend(_render_task(current))

    lines.extend(["", "Tasks:"])
    lines.extend(_render_task_list(tasks))

    lines.extend(["", "Recent Attempts:"])
    if not recent:
        lines.append("- none")
    else:
        for task in recent[:5]:
            lines.extend(_render_recent_task(task))

    lines.extend(["", "Paths:"])
    lines.append(f"- state: {Path(state.run_dir) / 'state.json'}")
    lines.append(f"- todo: {Path(state.run_dir) / 'TODO.md'}")
    lines.append(f"- prover: {Path(state.run_dir) / 'PROVER.md'}")
    lines.append(f"- issues: {Path(state.run_dir) / 'ISSUES.md'}")
    return "\n".join(lines)


def monitor_run(run_dir: Path, interval_seconds: float) -> None:
    while True:
        try:
            state = load_state(run_dir)
            state.run_dir = str(run_dir)
            snapshot = render_dashboard(state)
        except (FileNotFoundError, json.JSONDecodeError) as exc:
            logger.debug("failed to load state: %s", exc)
            snapshot = f"Waiting for state.json in {run_dir} ..."
        print("\x1bc" + snapshot, flush=True)
        time.sleep(interval_seconds)


def _render_task(task: StatementTask) -> list[str]:
    lines = [
        f"- {task.kind} {task.label} [{task.status}]",
        f"  attempts={task.attempts}, transport_failures={task.transport_failures}",
    ]
    if task.summary:
        lines.append(f"  summary: {task.summary}")
    if task.lean_path:
        lines.append(f"  lean: {task.lean_path}")
    excerpt = _single_line(task.source_excerpt, 220)
    if excerpt:
        lines.append(f"  excerpt: {excerpt}")
    return lines


def _render_recent_task(task: StatementTask) -> list[str]:
    lines = [f"- {task.label} [{task.status}] attempts={task.attempts}"]
    if task.attempt_history:
        lines.append(f"  last: {_single_line(task.attempt_history[-1], 220)}")
    return lines


def _render_task_list(tasks: list[StatementTask]) -> list[str]:
    return [_todo_line(task) for task in tasks]


def _todo_line(task: StatementTask) -> str:
    checked = "x" if task.status in {"verified", "formalized"} else " "
    prefix = "- [!]" if task.status == "proving" else f"- [{checked}]"
    if task.status == "verified":
        detail = "verified (verified)"
    elif task.status == "formalized":
        detail = "formalized (formalized)"
    elif task.status == "proving":
        detail = _single_line(task.summary, 100) if task.summary else "in progress"
    elif task.status == "retry":
        detail = _single_line(task.summary, 100) if task.summary else "retry"
    elif task.status == "weeks_scale":
        detail = _single_line(task.weeks_scale_reason or task.summary, 100) or "weeks-scale"
    else:
        detail = "pending"
    return f"{prefix} `{task.kind}` `{task.label}`: {detail}"


def _single_line(text: str | None, limit: int) -> str:
    if not text:
        return ""
    collapsed = " ".join(text.split())
    if len(collapsed) <= limit:
        return collapsed
    return collapsed[: limit - 3] + "..."



from __future__ import annotations

import json
import shutil
import subprocess
import time
from pathlib import Path

from .agents import (
    AgentInvocationError,
    AgentRunner,
    AgentTimeoutError,
    AgentTransportError,
)
from .arxiv import download_and_extract_arxiv_source
from .badges import (
    build_blob_url,
    build_local_file_target,
    copy_verified_badges_assets,
    ensure_verified_badges_repo,
    infer_repo_url_and_branch,
)
from .latex import (
    detect_main_tex,
    ensure_verified_badges_package,
    extract_statements_from_tree,
    insert_badge_for_task,
    set_badge_for_task,
)
from .models import RunState, StatementTask


def run_formalization(
    arxiv_id: str,
    base_dir: Path,
    local_paper_dir: Path | None,
    repo_url: str | None,
    branch: str | None,
    prover_backend: str,
    verifier_backend: str,
    badge_link_mode: str,
    verified_badges_repo: Path | None,
    agent_timeout_seconds: int | None,
    codex_model: str | None,
    claude_model: str | None,
    max_attempts_per_task: int | None,
    max_tasks: int | None,
) -> RunState:
    run_dir = base_dir / normalize_arxiv_id(arxiv_id)
    state = _load_or_initialize_state(
        arxiv_id=arxiv_id,
        run_dir=run_dir,
        local_paper_dir=local_paper_dir,
        repo_url=repo_url,
        branch=branch,
        prover_backend=prover_backend,
        verifier_backend=verifier_backend,
        badge_link_mode=badge_link_mode,
        verified_badges_repo=verified_badges_repo,
    )
    runner = AgentRunner(
        run_dir=Path(state.run_dir),
        paper_dir=Path(state.paper_dir),
        main_tex=Path(state.main_tex),
        root_module=state.root_module,
        prover_backend=state.prover_backend,
        verifier_backend=state.verifier_backend,
        agent_timeout_seconds=agent_timeout_seconds,
        codex_model=codex_model,
        claude_model=claude_model,
    )
    _persist_state(state)
    tasks_processed = 0
    while True:
        if max_tasks is not None and tasks_processed >= max_tasks:
            _persist_state(state)
            return state
        pending = [
            task
            for task in state.tasks
            if task.status in {"pending", "retry"}
            and (max_attempts_per_task is None or task.attempts < max_attempts_per_task)
        ]
        if not pending:
            _persist_state(state)
            return state
        task = pending[0]
        task.status = "proving"
        _persist_state(state)

        try:
            prover_result = runner.run_prover(task)
        except AgentTransportError as exc:
            task.status = "retry"
            task.summary = str(exc)
            task.last_error = str(exc)
            task.transport_failures += 1
            _append_attempt_history(
                task,
                f"Transport failure while starting attempt {task.attempts + 1}. {str(exc).strip()}",
            )
            tasks_processed += 1
            _persist_state(state)
            _sleep_after_transport_failure(task.transport_failures)
            continue
        except (AgentTimeoutError, AgentInvocationError) as exc:
            task.status = "retry"
            task.summary = str(exc)
            task.last_error = str(exc)
            _append_attempt_history(task, f"Attempt {task.attempts + 1}: agent failure. {str(exc).strip()}")
            tasks_processed += 1
            _persist_state(state)
            continue

        task.attempts += 1
        task.transport_failures = 0

        if prover_result.paper_issue:
            _append_issue(state, task, prover_result.paper_issue)

        missing_lean_reason = _missing_lean_path_reason(run_dir, prover_result.lean_path)
        if missing_lean_reason is not None:
            task.status = "retry"
            task.lean_path = prover_result.lean_path
            task.summary = missing_lean_reason
            task.last_error = missing_lean_reason
            _append_attempt_history(
                task,
                (
                    f"Attempt {task.attempts}: retry requested before verification. "
                    f"{missing_lean_reason} "
                    f"Prover: {prover_result.summary.strip()}"
                ),
            )
            tasks_processed += 1
            _persist_state(state)
            continue

        try:
            compile_ok, compile_output = _compile_candidate(run_dir, state.root_module, prover_result.lean_path)
            verifier_result = runner.run_verifier(task, prover_result, compile_ok, compile_output)
        except AgentTransportError as exc:
            task.status = "retry"
            task.lean_path = prover_result.lean_path
            task.summary = str(exc)
            task.last_error = str(exc)
            task.transport_failures += 1
            _append_attempt_history(
                task,
                f"Transport failure after substantive attempt {task.attempts}. {str(exc).strip()}",
            )
            tasks_processed += 1
            _persist_state(state)
            _sleep_after_transport_failure(task.transport_failures)
            continue
        except (AgentTimeoutError, AgentInvocationError) as exc:
            task.status = "retry"
            task.lean_path = prover_result.lean_path
            task.summary = str(exc)
            task.last_error = str(exc)
            _append_attempt_history(task, f"Attempt {task.attempts}: agent failure. {str(exc).strip()}")
            tasks_processed += 1
            _persist_state(state)
            continue

        if verifier_result.decision == "verified":
            task.transport_failures = 0
            _append_attempt_history(
                task,
                f"Attempt {task.attempts}: verifier accepted verified. compile_ok={compile_ok}. {verifier_result.summary.strip()}",
            )
            _mark_completed(state, task, prover_result.lean_path, "verified")
        elif verifier_result.decision == "formalized":
            task.transport_failures = 0
            _append_attempt_history(
                task,
                f"Attempt {task.attempts}: verifier accepted formalized. compile_ok={compile_ok}. {verifier_result.summary.strip()}",
            )
            _mark_completed(state, task, prover_result.lean_path, "formalized")
        elif verifier_result.decision == "weeks_scale" or prover_result.weeks_scale:
            task.status = "weeks_scale"
            task.summary = verifier_result.summary
            task.weeks_scale_reason = verifier_result.weeks_scale_reason or prover_result.weeks_scale_reason
            task.transport_failures = 0
            _append_attempt_history(
                task,
                f"Attempt {task.attempts}: classified weeks_scale. compile_ok={compile_ok}. {task.weeks_scale_reason or verifier_result.summary.strip()}",
            )
        else:
            task.status = "retry"
            task.lean_path = prover_result.lean_path
            task.summary = verifier_result.summary
            task.last_error = verifier_result.summary
            task.transport_failures = 0
            _append_attempt_history(
                task,
                (
                    f"Attempt {task.attempts}: retry requested. "
                    f"compile_ok={compile_ok}, exact_match={verifier_result.exact_match}, "
                    f"extra_hypotheses={verifier_result.extra_hypotheses}, lean_path={prover_result.lean_path or 'none'}. "
                    f"Verifier: {verifier_result.summary.strip()}. "
                    f"Prover: {prover_result.summary.strip()}"
                ),
            )

        tasks_processed += 1
        _persist_state(state)


def normalize_arxiv_id(arxiv_id: str) -> str:
    return arxiv_id.replace("/", "_").replace(":", "_")


def _load_or_initialize_state(
    arxiv_id: str,
    run_dir: Path,
    local_paper_dir: Path | None,
    repo_url: str | None,
    branch: str | None,
    prover_backend: str,
    verifier_backend: str,
    badge_link_mode: str,
    verified_badges_repo: Path | None,
) -> RunState:
    state_path = run_dir / "state.json"
    explicit_repo_target = repo_url is not None or branch is not None
    if state_path.exists():
        state = RunState.from_dict(json.loads(state_path.read_text(encoding="utf-8")))
        if Path(state.run_dir).resolve() != run_dir.resolve():
            old_run_dir = Path(state.run_dir)
            state.run_dir = str(run_dir)
            state.paper_dir = str(run_dir / Path(state.paper_dir).relative_to(old_run_dir))
            state.main_tex = str(run_dir / Path(state.main_tex).relative_to(old_run_dir))
        for task in state.tasks:
            if task.status == "proving":
                task.status = "retry"
                if task.last_error is None:
                    task.last_error = "Resuming after interrupted proving attempt."
                if task.summary is None:
                    task.summary = task.last_error
                _append_attempt_history(task, f"Supervisor resumed interrupted proving task `{task.label}` as retry.")
        if repo_url is not None:
            state.repo_url = repo_url
        if branch is not None:
            state.branch = branch
        if explicit_repo_target:
            state.explicit_repo_target = True
        state.prover_backend = prover_backend
        state.verifier_backend = verifier_backend
        state.badge_link_mode = badge_link_mode
        _synchronize_badges(state)
        _save_state(state)
        return state

    run_dir.mkdir(parents=True, exist_ok=True)
    paper_dir = run_dir / "paper"
    _populate_paper_dir(arxiv_id, paper_dir, local_paper_dir)
    tex_files = list((paper_dir).rglob("*.tex"))
    if not tex_files:
        raise RuntimeError(f"loaded paper source for {arxiv_id}, but found no .tex files")
    main_tex = detect_main_tex(tex_files)
    _initialize_lean_project(run_dir)
    root_module = _detect_root_module(run_dir)
    if root_module is None:
        raise RuntimeError("failed to detect generated Lean root module after `lake init`")
    badges_repo = ensure_verified_badges_repo(run_dir / ".cache", verified_badges_repo)
    copy_verified_badges_assets(badges_repo, main_tex.parent)
    ensure_verified_badges_package(main_tex)
    if repo_url is None or branch is None:
        inferred_repo_url, inferred_branch = infer_repo_url_and_branch(run_dir)
        repo_url = repo_url or inferred_repo_url
        branch = branch or inferred_branch

    tasks = [
        StatementTask(
            task_id=item.task_id,
            tex_path=item.tex_path,
            env_name=item.env_name,
            label=item.label,
            title=item.title,
            sequence_index=item.sequence_index,
            kind=item.kind,
            source_excerpt=item.source_excerpt,
        )
        for item in extract_statements_from_tree(paper_dir)
    ]
    if not tasks:
        raise RuntimeError("no tracked theorem-like environments found in the paper source")
    _write_instructions(run_dir)
    state = RunState(
        arxiv_id=arxiv_id,
        run_dir=str(run_dir),
        paper_dir=str(paper_dir),
        main_tex=str(main_tex),
        root_module=root_module,
        repo_url=repo_url,
        branch=branch,
        prover_backend=prover_backend,
        verifier_backend=verifier_backend,
        badge_link_mode=badge_link_mode,
        explicit_repo_target=explicit_repo_target,
        is_local_paper=local_paper_dir is not None,
        tasks=tasks,
    )
    _synchronize_badges(state)
    _save_state(state)
    return state


def _initialize_lean_project(run_dir: Path) -> None:
    if (run_dir / "lakefile.toml").exists() or (run_dir / "lakefile.lean").exists():
        return
    package_name = _lean_package_name_for_run(run_dir.name)
    subprocess.run(["lake", "init", package_name, "math"], cwd=run_dir, check=True)
    root_module = _detect_root_module(run_dir)
    if root_module is None:
        raise RuntimeError("failed to detect Lean root module after initialization")
    formalizations_dir = run_dir / root_module / "Formalizations"
    formalizations_dir.mkdir(exist_ok=True)
    for subdir in ["Definition", "Notation", "Lemma", "Proposition", "Corollary", "Theorem"]:
        (formalizations_dir / subdir).mkdir(parents=True, exist_ok=True)


def _compile_candidate(run_dir: Path, root_module: str | None, lean_path: str | None) -> tuple[bool, str]:
    missing_lean_reason = _missing_lean_path_reason(run_dir, lean_path)
    if missing_lean_reason is not None:
        return False, missing_lean_reason
    assert lean_path is not None
    lean_file = run_dir / lean_path
    text = lean_file.read_text(encoding="utf-8", errors="ignore")
    if "sorry" in text:
        return False, "Lean file still contains `sorry`."
    module_name = _module_name_for_file(run_dir, lean_file, root_module)
    if module_name is None:
        return False, f"Could not map {lean_path} to a Lean module in this package."
    result = subprocess.run(
        ["lake", "build", module_name],
        cwd=run_dir,
        text=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        check=False,
    )
    return result.returncode == 0, result.stdout.strip()


def _missing_lean_path_reason(run_dir: Path, lean_path: str | None) -> str | None:
    if lean_path is None:
        return "No Lean file path returned by prover."
    lean_file = run_dir / lean_path
    if not lean_file.exists():
        return f"Lean file does not exist: {lean_path}"
    return None


def _mark_completed(state: RunState, task: StatementTask, lean_path: str | None, badge_kind: str) -> None:
    task.status = badge_kind
    task.badge_kind = badge_kind
    task.lean_path = lean_path
    badge_url = _badge_target_for_task(state, lean_path)
    if badge_url:
        macro = "\\leanproof" if badge_kind == "verified" else "\\leanformalized"
        insert_badge_for_task(
            Path(state.paper_dir),
            task.task_id,
            task.tex_path,
            task.sequence_index,
            f"{macro}{{{badge_url}}}",
        )


def _persist_state(state: RunState) -> None:
    _synchronize_badges(state)
    _write_progress_files(state)
    _save_state(state)


def _write_progress_files(state: RunState) -> None:
    run_dir = Path(state.run_dir)
    todo_lines = ["# TODO", ""]
    for task in state.tasks:
        badge = f" ({task.badge_kind})" if task.badge_kind else ""
        todo_lines.append(f"- [{_checkbox(task.status)}] `{task.kind}` `{task.label}`: {task.status}{badge}")
    (run_dir / "TODO.md").write_text("\n".join(todo_lines) + "\n", encoding="utf-8")

    prover_lines = ["# PROVER", ""]
    active = [task for task in state.tasks if task.status == "proving"]
    if active:
        prover_lines.extend(
            f"- Working on `{task.label}` from `{task.tex_path}` (attempt {task.attempts})" for task in active
        )
    else:
        prover_lines.append("- Idle")
    (run_dir / "PROVER.md").write_text("\n".join(prover_lines) + "\n", encoding="utf-8")

    issues_path = run_dir / "ISSUES.md"
    if not issues_path.exists():
        issues_path.write_text("# ISSUES\n\n", encoding="utf-8")


def _append_issue(state: RunState, task: StatementTask, issue: str) -> None:
    issues_path = Path(state.run_dir) / "ISSUES.md"
    if not issues_path.exists():
        issues_path.write_text("# ISSUES\n\n", encoding="utf-8")
    with issues_path.open("a", encoding="utf-8") as handle:
        handle.write(f"- `{task.label}`: {issue.strip()}\n")


def _write_instructions(run_dir: Path) -> None:
    instructions = """# Role of Agents

## Prover

Use Codex as a prover. Formalize statements from `paper/` in Lean 4. Keep `PROVER.md` current while working. Create one readable Lean file per statement where practical. Fix linter warnings in touched Lean files.

## Verifier

Use the verifier to check that each Lean formalization matches the paper exactly and that there are no hidden hypotheses. If a proof is complete, the supervisor may add a verified badge. If only the statement is exact, the supervisor may add a formalized badge.

## General Instructions

Do not stop a task early except for genuine paper issues or when the remaining work would honestly take LLM-weeks. Put paper issues into `ISSUES.md`. Keep `TODO.md` current.
"""
    (run_dir / "INSTRUCTIONS.md").write_text(instructions, encoding="utf-8")


def _save_state(state: RunState) -> None:
    tmp_path = state.state_path.with_suffix(".tmp")
    tmp_path.write_text(json.dumps(state.to_dict(), indent=2, sort_keys=True), encoding="utf-8")
    tmp_path.replace(state.state_path)


def _append_attempt_history(task: StatementTask, entry: str) -> None:
    task.attempt_history.append(entry.strip())
    if len(task.attempt_history) > 12:
        task.attempt_history[:] = task.attempt_history[-12:]


def _sleep_after_transport_failure(failures: int) -> None:
    delay_seconds = min(60, 5 * (2 ** max(0, failures - 1)))
    time.sleep(delay_seconds)


def _checkbox(status: str) -> str:
    return "x" if status in {"verified", "formalized", "weeks_scale"} else " "


def _badge_target_for_task(state: RunState, lean_path: str | None) -> str | None:
    if lean_path is None:
        return None
    absolute_path = Path(state.run_dir) / lean_path
    mode = state.badge_link_mode
    if mode == "local":
        return build_local_file_target(Path(state.paper_dir), absolute_path)
    if mode == "github":
        return build_blob_url(state.repo_url, state.branch, absolute_path)

    # auto
    if state.explicit_repo_target and state.repo_url and state.branch:
        return build_blob_url(state.repo_url, state.branch, absolute_path)
    return build_local_file_target(Path(state.paper_dir), absolute_path)


def _synchronize_badges(state: RunState) -> None:
    paper_dir = Path(state.paper_dir)
    for task in state.tasks:
        badge_kind = None
        if task.status in {"verified", "formalized"}:
            badge_kind = task.badge_kind or task.status
        if badge_kind is None:
            set_badge_for_task(paper_dir, task.task_id, task.tex_path, task.sequence_index, None)
            continue
        badge_target = _badge_target_for_task(state, task.lean_path)
        if badge_target is None:
            set_badge_for_task(paper_dir, task.task_id, task.tex_path, task.sequence_index, None)
            continue
        macro = "\\leanproof" if badge_kind == "verified" else "\\leanformalized"
        set_badge_for_task(
            paper_dir,
            task.task_id,
            task.tex_path,
            task.sequence_index,
            f"{macro}{{{badge_target}}}",
        )


def _lean_package_name_for_run(run_name: str) -> str:
    sanitized = "".join(ch if ch.isalnum() else "_" for ch in run_name)
    while "__" in sanitized:
        sanitized = sanitized.replace("__", "_")
    sanitized = sanitized.strip("_")
    if not sanitized:
        sanitized = "ArxivPaper"
    if not sanitized[0].isalpha():
        sanitized = f"Arxiv_{sanitized}"
    return sanitized


def _detect_root_module(run_dir: Path) -> str | None:
    candidates = sorted(
        path.parent.name for path in run_dir.glob("*/Basic.lean") if path.parent.name not in {"paper", ".lake"}
    )
    return candidates[0] if candidates else None


def _module_name_for_file(run_dir: Path, lean_file: Path, root_module: str | None) -> str | None:
    try:
        relative = lean_file.resolve().relative_to(run_dir.resolve())
    except ValueError:
        return None
    if relative.suffix != ".lean":
        return None
    parts = list(relative.with_suffix("").parts)
    if root_module is not None and (not parts or parts[0] != root_module):
        return None
    return ".".join(parts)


def _populate_paper_dir(arxiv_id: str, paper_dir: Path, local_paper_dir: Path | None) -> None:
    if local_paper_dir is None:
        download_and_extract_arxiv_source(arxiv_id, paper_dir)
        return
    if not local_paper_dir.exists():
        raise RuntimeError(f"local paper directory does not exist: {local_paper_dir}")
    if not local_paper_dir.is_dir():
        raise RuntimeError(f"local paper path is not a directory: {local_paper_dir}")
    shutil.copytree(local_paper_dir, paper_dir, dirs_exist_ok=True)



from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from paper_badger import cli


def test_main_defaults_to_run_mode_for_legacy_invocation(tmp_path: Path) -> None:
    with (
        patch("paper_badger.cli.run_formalization") as run_formalization,
        patch(
            "sys.argv",
            ["paper_badger.cli", "2401.01234", "--runs-dir", str(tmp_path)],
        ),
    ):
        run_formalization.return_value = type(
            "State",
            (),
            {"arxiv_id": "2401.01234", "tasks": []},
        )()
        cli.main()

    assert run_formalization.call_args.kwargs["arxiv_id"] == "2401.01234"
    assert run_formalization.call_args.kwargs["base_dir"] == tmp_path.resolve()
    assert run_formalization.call_args.kwargs["prover_backend"] == "codex"
    assert run_formalization.call_args.kwargs["verifier_backend"] == "claude"


def test_main_allows_explicit_role_backend_overrides(tmp_path: Path) -> None:
    with (
        patch("paper_badger.cli.run_formalization") as run_formalization,
        patch(
            "sys.argv",
            [
                "paper_badger.cli",
                "2401.01234",
                "--runs-dir",
                str(tmp_path),
                "--prover-backend",
                "claude",
                "--verifier-backend",
                "codex",
            ],
        ),
    ):
        run_formalization.return_value = type(
            "State",
            (),
            {"arxiv_id": "2401.01234", "tasks": []},
        )()
        cli.main()

    assert run_formalization.call_args.kwargs["prover_backend"] == "claude"
    assert run_formalization.call_args.kwargs["verifier_backend"] == "codex"


def test_main_monitor_once_renders_snapshot(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    run_dir = tmp_path / "paper-id"
    run_dir.mkdir()

    with (
        patch("paper_badger.cli.render_dashboard", return_value="dashboard"),
        patch("paper_badger.cli.load_state", return_value=MagicMock()),
        patch(
            "sys.argv",
            ["paper_badger.cli", "monitor", "paper-id", "--runs-dir", str(tmp_path), "--once"],
        ),
    ):
        cli.main()

    captured = capsys.readouterr()
    assert captured.out.strip() == "dashboard"



from pathlib import Path

from paper_badger.latex import (
    detect_main_tex,
    ensure_verified_badges_package,
    extract_statements_from_tree,
    insert_badge_for_task,
    locate_statement,
    set_badge_for_task,
    statement_context_summary,
)


def test_extract_statements_and_insert_badge(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\begin{document}
\begin{definition}[Widget]
\label{def:widget}
A widget is a gadget.
\end{definition}
\begin{theorem}[Main result]
\label{thm:main}
Every widget is useful.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)

    assert [statement.label for statement in statements] == ["def:widget", "thm:main"]

    theorem = next(statement for statement in statements if statement.label == "thm:main")
    inserted = insert_badge_for_task(
        paper_dir,
        theorem.task_id,
        theorem.tex_path,
        theorem.sequence_index,
        r"\leanproof{https://example.com/proof}",
    )

    assert inserted
    updated = tex.read_text(encoding="utf-8")
    assert r"\begin{theorem}[Main result \leanproof{https://example.com/proof}]" in updated


def test_usepackage_insertion_and_main_tex_detection(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    helper = paper_dir / "helper.tex"
    helper.write_text(r"\begin{lemma}Hi\end{lemma}", encoding="utf-8")
    main = paper_dir / "main.tex"
    main.write_text(
        r"""
\documentclass{article}
\usepackage{amsmath}
\begin{document}
Hello
\end{document}
""".strip(),
        encoding="utf-8",
    )

    detected = detect_main_tex([helper, main])
    changed = ensure_verified_badges_package(main)
    changed_again = ensure_verified_badges_package(main)

    assert detected == main
    assert changed
    assert not changed_again
    assert r"\usepackage{verified-badges}" in main.read_text(encoding="utf-8")


def test_extract_ignores_preamble_and_supports_uppercase_envs(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\newenvironment{remark}[1]{\begin{Rem}\label{R:#1}}{\end{Rem}}
\begin{document}
\begin{Prop}\label{P:main}
Useful proposition.
\end{Prop}
\begin{Rem}\label{R:obs}
Helpful observation.
\end{Rem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)

    assert [(s.kind, s.label) for s in statements] == [("proposition", "P:main")]


def test_extract_skips_commented_out_statement_bodies_and_prioritizes_labeled_items(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\begin{document}
\begin{definition}
Visible unlabeled definition.
\end{definition}
\begin{lemma}
\label{lem:commented}
% This whole lemma statement is commented out.
% It should not become a tracked task.
\end{lemma}
\begin{theorem}
\label{thm:labeled}
Visible labeled theorem.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)

    assert [statement.label for statement in statements] == ["thm:labeled", "synthetic:definition:definition-0"]


def test_locate_and_context_work_with_sparse_sequence_indices(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\title{Sparse Index Note}
\begin{document}
\begin{lemma}
\label{lem:commented}
% This one is commented out.
\end{lemma}
\begin{theorem}
\label{thm:kept}
Visible theorem.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)
    theorem = statements[0]
    located = locate_statement(paper_dir, theorem.task_id, theorem.tex_path, theorem.sequence_index)
    context = statement_context_summary(paper_dir, theorem.tex_path, theorem.sequence_index)

    assert theorem.sequence_index == 1
    assert located.extracted.task_id == theorem.task_id
    assert "Paper title: Sparse Index Note" in context


def test_insert_badge_preserves_preamble_text(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\usepackage{amsmath}
\begin{document}
\begin{definition}[Widget]
\label{def:widget}
A widget is a gadget.
\end{definition}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)
    definition = statements[0]
    inserted = insert_badge_for_task(
        paper_dir,
        definition.task_id,
        definition.tex_path,
        definition.sequence_index,
        r"\leanproof{https://example.com/proof}",
    )

    updated = tex.read_text(encoding="utf-8")

    assert inserted
    assert updated.startswith("\\documentclass{article}\n\\usepackage{amsmath}\n")
    assert r"\begin{definition}[Widget \leanproof{https://example.com/proof}]" in updated


def test_set_badge_for_task_can_replace_and_remove_badges(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\begin{document}
\begin{theorem}[Main result \leanproof{old/path.lean}]
\label{thm:main}
Every widget is useful.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    statements = extract_statements_from_tree(paper_dir)
    theorem = statements[0]

    replaced = set_badge_for_task(
        paper_dir,
        theorem.task_id,
        theorem.tex_path,
        theorem.sequence_index,
        r"\leanformalized{new/path.lean}",
    )
    removed = set_badge_for_task(
        paper_dir,
        theorem.task_id,
        theorem.tex_path,
        theorem.sequence_index,
        None,
    )

    updated = tex.read_text(encoding="utf-8")

    assert replaced
    assert removed
    assert r"\leanproof{old/path.lean}" not in updated
    assert r"\leanformalized{new/path.lean}" not in updated
    assert r"\begin{theorem}[Main result]" in updated


def test_statement_context_summary_includes_title_and_previous_statements(tmp_path: Path) -> None:
    paper_dir = tmp_path / "paper"
    paper_dir.mkdir()
    tex = paper_dir / "paper.tex"
    tex.write_text(
        r"""
\documentclass{article}
\title{A Note About Natural Numbers}
\begin{document}
\begin{definition}[Even natural number]
\label{def:even}
A natural number $n$ is even if there exists a natural number $k$ such that $n = 2k$.
\end{definition}
\begin{theorem}[Main]
\label{thm:main}
If $m$ and $n$ are even, then $m+n$ is even.
\end{theorem}
\end{document}
""".strip(),
        encoding="utf-8",
    )

    context = statement_context_summary(paper_dir, "paper.tex", 1)

    assert "Paper title: A Note About Natural Numbers" in context
    assert "definition `def:even`" in context
    assert "A natural number $n$ is even" in context



from __future__ import annotations

import subprocess
from pathlib import Path  # noqa: TC003
from typing import Any
from unittest.mock import patch

from paper_badger.agents import AgentRunner


def test_detects_transport_disconnect_output() -> None:
    output = """
2026-03-21 WARN codex_core::codex: stream disconnected before completion
Reconnecting... 1/5 (stream disconnected before completion: error sending request for url (https://chatgpt.com/backend-api/codex/responses))
""".strip()

    assert AgentRunner._looks_like_transport_error(output)


def test_ignores_non_transport_output() -> None:
    output = "error: Lean elaboration failed because theorem statement is ill-typed"

    assert not AgentRunner._looks_like_transport_error(output)


def test_extract_json_object_from_mixed_output() -> None:
    output = """
Some progress text
{"status":"verified_candidate","summary":"ok","lean_path":"Paper/Test.lean","proof_complete":true,"weeks_scale":false,"weeks_scale_reason":null,"paper_issue":null}
""".strip()

    payload = AgentRunner._extract_json_object(output)

    assert payload["status"] == "verified_candidate"
    assert payload["lean_path"] == "Paper/Test.lean"


def test_run_claude_prover_uses_run_dir_cwd_and_extracts_json(tmp_path: Path) -> None:
    runner = AgentRunner(
        run_dir=tmp_path,
        paper_dir=tmp_path / "paper",
        main_tex=tmp_path / "paper" / "main.tex",
        root_module="Paper",
        prover_backend="claude",
        verifier_backend="claude",
        agent_timeout_seconds=30,
    )

    calls: list[dict[str, Any]] = []

    def fake_run(cmd: list[str], **kwargs: object) -> subprocess.CompletedProcess[str]:
        calls.append({"cmd": cmd, "cwd": kwargs.get("cwd")})
        return subprocess.CompletedProcess(
            args=cmd,
            returncode=0,
            stdout=(
                "Some progress text\n"
                '{"status":"verified_candidate","summary":"ok","lean_path":"Paper/Test.lean",'
                '"proof_complete":true,"weeks_scale":false,"weeks_scale_reason":null,"paper_issue":null}\n'
            ),
        )

    with (
        patch("paper_badger.agents.shutil.which", return_value="/usr/bin/claude"),
        patch("paper_badger.agents.subprocess.run", side_effect=fake_run),
    ):
        payload = runner._run_claude("Prove something.", {"type": "object"}, role_name="prover", cwd=tmp_path)

    assert payload["lean_path"] == "Paper/Test.lean"
    assert calls[0]["cwd"] == tmp_path
    assert calls[0]["cmd"][:2] == ["claude", "-p"]
